%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS query.ql,v 1.20 1993/12/10 02:47:57 marku Exp



/*
** Locate a fact in database with given extension by searching,
** the current and ancestor theories in an unspecified order.
*/


find_up(Name, Params, Body) :- 
	ancestor(PT),
	clause([Name,PT] : Params, Body).

find_at(Name, CD, Params, Body) :-
	clause([Name, CD] : Params, Body).



/*
** Same as find_up except that guard (Body) is executed
*/

get_up(Name, Params) :- ancestor(PT), ([Name,PT]: Params).

get_up_dir(Name, Params, PT) :- ancestor(PT), ([Name,PT]: Params).

get_at(Name, CD, Params) :- ([Name, CD] : Params).


%%
%% Perform search for an axiom or theorem in the current 
%% directory, then in each of those above.
%%
%% This search strategy may no longer be terribly useful, although
%% having a single call for both rule types remains useful in any 
%% database structure.
%%

is_theorem(Name, Form) :-
	decode_rule(Name, CD, RuleName), 
	look_for_theorem(CD:RuleName, Form),
	rule_at2(RuleName, Form, CD).


rule_at2(Name, Form, CD) :- ([axiom, CD] : [Name, Form]).
rule_at2(Name, Form, CD) :- ([theorem, CD] : [Name, Form]).
rule_at2(Name, Form, CD) :- ([lemma, CD] : [Name, Form]).


/*
** Find rule and guard at a given directory
*/


find_rule_at2(Name,Form,Body,CD) :- clause(([theorem,CD] : [Name, Form]), Body).
find_rule_at2(Name,Form,Body,CD) :- clause(([axiom,CD]   : [Name, Form]), Body).
find_rule_at2(Name,Form,Body,CD) :- clause(([lemma,CD]   : [Name, Form]), Body).

/*
** Restrict search for a probably instantiated theorem by
** examining the outermost function symbol
*/

look_for_theorem(PT:Name, Form) :-
        (nonvar(PT) ; var(Form)), !,
        ancestor(PT).

look_for_theorem(PT:Name, Form) :-
        functor(Form, Op, Arity), !,
	(atom(Op) ->
        	inherited_symbol(Op, Arity, Children),
        	member(PT, Children)
	;
		ancestor(PT)).

look_for_theorem(PT:Name, Form) :-
        quantify(Form, Q, _, _), !,
	(atom(Q) ->
        	inherited_symbol(Q, quant, Children),
        	member(PT, Children)
	;
		ancestor(PT)).

look_for_theorem(PT:Name, Form) :-
        ancestor(PT).


/*
** Find rule and guard at some parent directory
*/


find_rule(Name, Form, Body) :-
	decode_rule(Name, CD, RuleName), 
	look_for_theorem(CD:RuleName, Form),
	find_rule_at2(RuleName, Form, Body, CD).


/*
** Allow reuse of theorem/axiom names only in separate directories.
*/

available(Name) :-
	current_dir(CD), 
	check_or_writelist(
		\+  (
			find_rule_at2(Name, _, _, CD)
		    ;
		        get_compiled_rule(CD, Name, _, _)
		    ),
		['There is already a rule in ''', CD,''' called ''',
		  Name, '''.', nl]).

/*
** Decode the Dir:Name convention
*/

decode_rule(Dir:Name, CD, Name) :- !, CD = Dir.
decode_rule(Name, CD, Name).


%% stronger(+Reln1, +Reln2)
%%
%%  Succeeds iff Reln1 is stronger than (or equal to) Reln2
%%  and both relations are known to be preorders or rev/1 applied
%%  once or twice to a known preorder.

stronger(A,B) :-
	equal_to(A,A2),
	equal_to(B,B2),
	stronger_than(A2, B2),
	!.

%% equal_to(?Reln0, ?Reln)
%%
%%  This simplifies Reln0 down to a near canonical form,
%%  by removing rev/1 wrappers where possible.
%%  The relations are not required to be preorders.

equal_to(A, B) :-
	nonvar(A),
	(
	    A = rev(rev(B)),
	    atom(B)
	;
	    equal_preorder(A, B)
	),
	!.

equal_to(A, A).


