%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS theory_dag.ql,v 1.15 1993/12/10 02:23:19 marku Exp



/*
** Manage the multiple inheritance theory structure.
*/ 


add_theory(Name) :- 
    check_atom(Name),
    Name \== global,
    \+  is_theory_name(Name),
    assert_thy(Name, status, [open]),
    assert_global(ancestor, [Name, []]),
    assert(loaded(Name)), !,
    cd(Name).

/*
** - If a theory inherits an operator from two places, then it must
**   have only one point of origin.
*/


inherit_theory(Parent) :-
    current_dir(CD),
    nonvar(Parent),
    check_or_writelist( 
	\+  ancestor(Parent), 
        [CD, ' already inherits ', Parent, nl]),
    check_or_writelist( 
	check_compatible(CD, Parent),
	[CD, ' and ', Parent, ' inherit (possibly) clashing definitions.',nl]),
    check_or_writelist(
        (ancestor : [Parent, PList]),
	[Parent, ' is not defined as a theory.', nl]),
    (ancestor : [CD, CList]),
    merge_atoms([Parent], PList,   Inter),
    merge_atoms(Inter,    CList,   FinalList),
    check_or_writelist(
	check_circularities(CD, FinalList),
	['This would result in a circularity in the theory structure.',nl]),
    retract(ancestor : [CD, _]),
    assert_global(ancestor, [CD, FinalList]),
    assert_global(parent, [CD, Parent]),
    load_ancestors(CD),
    mk_child_lists,
    mk_inherited_symbols,
    mk_preorder_info.


merge_atoms([], L, L).
merge_atoms([A|B], L, D) :- member(A, L), !, merge_atoms(B, L, D).
merge_atoms([A|B], L, [A|D]) :- merge_atoms(B, L, D).


check_circularities(E, L) :-
	\+  member(E, L).

check_incompatible_theories(A, B) :-
	A \== B,
	once((
	        (incompatible_theories : [B, A])
	    ;
		(incompatible_theories : [A, B])
	    )).
	
check_compatible(T1, T2) :-
	\+  (ancestor_of(T1, A1),
	     ancestor_of(T2, A2),
	     check_incompatible_theories(A1, A2)).
	

check_co_inherited(T, T).
check_co_inherited(T1, T2) :-
	(ancestor : [T, L]),
	member(T1, [T|L]),
	member(T2, [T|L]).


/*
**  Heritage functions.  Note that ancestor_of is reflexive.
*/

ancestor_of(CD, PT)     :- (ancestor : [CD, Parents]), member(PT, [CD|Parents]).
strict_ancestor(CD, PT) :- (ancestor : [CD, Parents]), member(PT, Parents).


/*
** If a symbol has been declared elsewhere, it must have the
** same precedence.
**
** None of a quant, obvar, default/pre/post/in - fix functions may
** share a name.
**
** A symbol cannot be declared twice (with the same prec, arity)
** in the one directory
*/



parse_condition(default, A1,  default, A2).
parse_condition( P, A,  P, A).
 	

check_compatibility(Th, Prec, OpInfo, Name) :-
	\+  (
	    (symbol : [ThA, PrecA, OpInfoA, Name]),
	    once((
	            (\+ parse_condition(Prec, OpInfo, PrecA, OpInfoA))
		;
		    Prec = PrecA, 
		    OpInfo = OpInfoA,
		    \+  check_incompatible_theories(Th, ThA),
		    once(check_co_inherited(Th, ThA))
		))
	    ).


update_compatibility(Th, P, A, Name) :-
	(symbol : [T2, P, A, Name]),
	\+  check_incompatible_theories(Th, T2),
	assert_global(incompatible_theories, [Th, T2]),
	fail.

update_compatibility(Th, P, A, Name).
	


claim_symbol(Prec, OpInfo, Name) :-
	check_atomic(Name),
	nonvar(Prec),
	nonvar(OpInfo),
	check_or_writelist(
		valid(Prec, OpInfo),
		['Invalid precedence/opinfo',nl]),
	once((
	        current_dir(CD)
	    ;
	        CD = global
	    )),
	check_compatibility(CD, Prec, OpInfo, Name),
	update_compatibility(CD, Prec, OpInfo, Name),
        localop(Prec, OpInfo, Name),
        assert_global(symbol, [CD, Prec, OpInfo, Name]),
        add_inherited_symbol(Name, OpInfo, CD, true).


valid(default, I) :- integer(I).
valid(obvar, obvar).
valid(P, A) :- integer(P), check_atom(A).


activate_symbols :-
	\+  ((symbol : [_, Prec, Arity, Name]),
	     \+  localop(Prec, Arity, Name)). 

/*
**  Is Name a theory?
*/

is_theory_name(global). 
is_theory_name(Name) :- 
	(ancestor : [Name, _]).


/*
** Change Theory
*/

cd(Name) :- 
	current_dir(Name), !.

cd(Name) :-
	check_or_writelist(
		\+  in_proof, 
		['Can''t change theories during a proof session',nl]),
	nonvar(Name),
	load_ancestors(Name),
	abolish(current_dir,1),
	assert(current_dir(Name)),
	mk_child_lists,
	mk_inherited_symbols,
	mk_preorder_info.


load_ancestors(Name) :-
	abolish(ancestor,1),
	\+  (ancestor_of(Name, A),
	     \+  assertz(ancestor(A))),
	load_ancestors_2(Name).

load_ancestors_2(Name) :-
	\+  ((parent : [Name, T]),
	     \+  load_ancestors_2(T)),
	(\+ loaded(Name) ->
	    load_dir(Name)
	;
	    true
	).


load_dir(Name) :-
	assert(loading(Name)),
	file_name(Name, Data),
	do_consult(Data),
	load_tactics(Name),
	retract(loading(Name)).

load_tactics(Name) :-
       ([tactic_file,Name]:[File]),
       do_consult(File),
       fail.

load_tactics(Name).


load_global_tactics :-
	\+  ((tactic_file : [File]),
	     \+  do_consult(File)).
	
	

/* 
** mark current theory as closed 
*/

close_theory :-
    current_dir(CD),
    retract_thy(CD, status, [_]),
    assert_thy(CD, status, [closed]).


/* 
**  check that a given theory has open status 
*/

cd_open(CD) :-
        cd(CD),
        check_open(CD).

check_open :-
	current_dir(CD),
	check_open(CD).

check_open(Theory) :-
    check_or_writelist(
	([status,Theory]:[open]), 
    	['Error: Theory is not open.',nl]).

/*
** Set the default interactive tactic
*/

set_prover(T) :-
    retractall(interactive_tactic : [_]), 
    assert_global(interactive_tactic, [T]).


/* 
** Make the necessary parser/obvar declarations for a symbol
*/

localop(default, _,     _).
localop(obvar,   _,     Name) :- object_var(Name).
localop(Prec,    Arity, Name) :- integer(Prec), op(Prec, Arity, Name).
	

op_arity(xf, 1).
op_arity(yf, 1).
op_arity(fy, 1).
op_arity(fx, 1).
op_arity(xfy, 2).
op_arity(yfx, 2).
op_arity(xfx, 2).
op_arity(A, A).


rev_tr(A, B) :- rev_tr(A, [], B).

rev_tr([], B, B).
rev_tr([A|R], In, B) :-
	rev_tr(R, [A|In], B).

mk_child_lists :-
	abolish(children, 2),
	current_dir(C),
	assert(children(C, [C])),
	(ancestor : [C, All]), 
	\+ (member(A, All),
	    mk_child_list(All, A, [C], Out),  
	    rev_tr([A|Out], AOut),
	    \+ assert(children(A, AOut))).

mk_child_list([], A, In, In).
mk_child_list([T | Ts], A, In, [T | Out]) :-
	(ancestor : [T, TA]),
	member(A, TA), !,
	mk_child_list(Ts, A, In, Out).
mk_child_list([_ | Ts], A, In, Out) :-
	mk_child_list(Ts, A, In, Out).
	 
	
	
intersect([], A, []).
intersect([E | B], A, [E | C]) :-
        member(E, A), !,
        intersect(B, A, C).
intersect([E | B], A, C) :-
        intersect(B, A, C).


add_inherited_symbol(Name, Op, PT, Body) :-
	once(op_arity(Op, Arity)),
	children(PT, CPT),
	assert((inherited_symbol(Name, Arity, CPT) :- Body)).


mk_inherited_symbols :-
	abolish(inherited_symbol,3),
	\+  (ancestor(PT),
	     clause((symbol : [PT, _, Op, Name]), Body),
	     \+  add_inherited_symbol(Name, Op, PT, Body)).



mk_preorder_info :-
	abolish(preorder, 1),
	abolish(equivalence, 1),

	\+ (get_up(preorder, [P]),
	    \+ assert(preorder(P))),

        \+ (get_up(equivalence, [E]),
            \+ assert(equivalence(E))),

	abolish(stronger_than,2),
	abolish(equal_preorder,2),

	\+  (preorder(P),
	     mk_strength_info(P,P),
	     (equivalence(P) 
		->
		assert(equal_preorder(rev(P),P))
		;
		(get_up(reverse_of, [P, RP])
			->
			assert(equal_preorder(rev(P), RP))
			; 
			true
		)	
	     ),
	     fail).


mk_strength_info(P, Q) :-
	stronger_than(P, Q), !.

mk_strength_info(P, Q) :-
	(P == Q
	   ->
		assert(stronger_than(P,P)),
		assert(stronger_than(rev(P),rev(P)))
	   ;
		assert(stronger_than(P, Q)),
		assert(stronger_than(rev(P), rev(Q))),
		(once((equivalence(P) ; equivalence(Q))) 
		 -> 
			assert(stronger_than(P, rev(Q))),
			assert(stronger_than(rev(P), Q))
		; 
		true)
	),
	\+  (get_up(stronger_than, [Q, R]),
	     \+  mk_strength_info(P, R)), !.
	     
