%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS hyps.ql,v 1.14 1993/12/10 02:23:45 marku Exp


?- op(999, fx,  (h)).
?- op(810, xfy, (and)).


/*
** Dependancies
**
**	Metavar (frozen)   -> conjecture
**	Ground term	   -> Proven hypothesis
**	Term with metavars -> Partially proven hypothesis 
**
** To test whether a hypothesis depends on an unproven conjecture (this 
** is the only dependancy of interest to the prover), test unifiability of the
** two justifications.  
**
** It is tantilizingly not quite possible to unify the dependancy and
** justification concepts.  It is I think more profitiable to view dependancies
** as a highly processed form of the justification information.
**
** Special attention will have to be paid to this information on rerun, as
** discharged conjectures may reference hypotheses introduced after
** the conjecture was raised.
**
*/


/*
** Add a new hypothesis
*/

add_hyp(Term, Next, Just, Dep) :-
	ipLookup(hyplist, Hyps),
	ipLookup(hypnr, No),
    	Next is No + 1,
	ipSet(hypnr, Next),
	ipSet(hyplist, [(Next, Term, _, Just, Dep) | Hyps]).


add_hyps([], _, _).
add_hyps([A|B], J, D) :-
	add_hyp(A, _, J, D), 
	add_hyps(B, J, D).

/*
** split hyps along and's
**
** equivalent to :
**      safe(h_trans(Hyp, imp_conj_1)),
**      h_trans(Hyp, imp_conj_2)
*/


split_hyp(Hyp) :-
	delete_hyp(Hyp, (N, A and B, _, _, Dep)), 
	add_hyp(A, _, split_left(N), Dep),
	add_hyp(B, _, split_right(N), Dep).

	
split_left(Hyp, No) :-
	get_hyp(Hyp, (N, A and B, _, _, Dep)), 
	add_hyp(A, No, split_left(N), Dep).
	
split_right(Hyp, No) :-
	get_hyp(Hyp, (N, A and B, _, _, Dep)), 
	add_hyp(B, No, split_right(N), Dep).


/*
** Decode a hypothesis name
*/

decode_hyp(A, 		      (A, _, _, _, _)) :- var(A), !.
decode_hyp(h N,               (N, _, _, _, _)) :- integer(N).
decode_hyp(h T,               (_, T, _, _, _)) :- !.
decode_hyp(hypotheses : B, X)                  :- !, decode_hyp(B, X).
decode_hyp((A, B, C, D, _),   (A, B, C, D, _)).
decode_hyp(N,                 (N, _, _, _, _)) :- integer(N).
	

/*
**  Internal use only!!!
*/

get_hyp(Name, Hyp) :-
	ipLookup(hyplist, HL),
	get_hyp(Name, Hyp, HL).

get_hyp(Name, (No, T, Del, J, Dep), HL) :-
	decode_hyp(Name, (No, T, Del, J, Dep)),
	member((No, T, Del, J, Dep), HL),
	var(Del).


/*
** delete_hyp/n    won't delete conjectures, although it should be
**		    possible for un-used conjectures with a little work,
**                  and for used conjectures and their dependants with
**		    quite a bit of work. 
**		    
*/

delete_hyp(Name)      :- ipLookup(hyplist, HL), delete_hyp(Name, _, HL).
delete_hyp(Name, Hyp) :- ipLookup(hyplist, HL), delete_hyp(Name, Hyp, HL).

delete_hyp(Name, (No, T, Del, J, Dep), HL) :- 
	get_hyp(Name, (No, T, Del, J, Dep), HL),
	(\+ (var(Dep), 
	     functor(J, conjecture, _)),
	 ipLookup(safe_hyps, false)
	->
	    Del = deleted
	;
	    true
	).

/*
**  Simple operations on hypothesis structure
*/

number_of_hyps(No) :- ipLookup(hypnr, No).

is_hyp(No, T)      :- get_hyp(_,   (No, T, _, _, _)), incomplete_retry_delays.
is_hyp(Hyp, No, T) :- get_hyp(Hyp, (No, T, _, _, _)), incomplete_retry_delays.


/*
**  Find true things during a proof
*/


get_th_or_hyp(Hyp,  No,             T, true, Dep) :- 
	get_hyp(Hyp, (No, T, _, _, Dep)).

get_th_or_hyp(Rule, CD:RuleName,    T, Body, Dep) :- 
	decode_rule(Rule, CD, RuleName),	
        look_for_theorem(CD:RuleName, T),
        (
	    find_rule_at2(RuleName, T, Body, CD),
	    Dep = []
	;
	    once(eval_compiled_rule(CD,RuleName,_,_,T,Dep)),
	    Body = true
	).

get_th_or_hyp(Conj,  No,           T, true, Dep) :-
	nonvar(Conj),
	Conj = c(No),
	var(No),
        '$freeze2'([Dep, Just, W]),                  %% Avoid syntax checking
        add_hyp(T, No, conjecture(W, Just), Dep), !. %% in conjecture/3

/*
** Perform 'Code' with hypothesis deletion turned off.
**
** safe(delete_hyp(_)) has no effect on the proof state
*/

safe(Code) :-
	ipLookup(safe_hyps, SH),
	ipSet(safe_hyps, true),
	!,
	Code,
	!,
	ipSet(safe_hyps, SH).


/*
** Add a conjecture to the current window
*/

conjecture(Term) :- conjecture(Term, _, _).
conjecture(Term, N) :- conjecture(Term, N, _).

conjecture(Term, N, Dep) :-
	check_term(Term),
	assimilate_new_vars([Term]),
	'$freeze2'([Dep, Just, W]),
	add_hyp(Term, N, conjecture(W, Just), Dep).
