%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS newuse.ql,v 1.11 1993/12/10 02:18:56 marku Exp


%% This file implements the use, fwd and print commands.
%% =====================================================
%%
%% All these commands use search/2 to find and
%% instantiate an appropriate axiom/theorem/hypothesis.

?- object_var('x').

:- declare(print/1).
:- declare(print/2).

%% use(+Query, ?Goal, ?ConjectureList)
%%
%%  ConjectureList is typically uninstantiated or an empty list
%%  when use is called.  An empty list prevents use generating any
%%  conjectures.  If use does generate any conjectures, they are
%%  returned in ConjectureList.  Eg. [c(4), c(5)].

use(Q)      :- use(Q, _, []).
use(Q,G)    :- use(Q, G, []).

use(Query, GoalSpec, ConjList) :-
    srch_expand_term(GoalSpec, Goal),
    ((var(Goal),(\+ frozen(Goal)) ->
	Msg = ['use(', Name ,', _' | MsgTail])
    ;
	(Msg = ['use(', Name, ', ', Goal | MsgTail])
    ),
    approx_copy(Goal, SearchGoal),
    is_reln(Reln),
    thawing_new_vars((
	search(Query and input=focus and reln=Reln and output=SearchGoal,
	       Ans),
	search_answer(body,  Ans, Body),
	with_frozen_nfi_test(on,Body)
        )),
    search_answer(name,      Ans, Name),
    search_answer(direction, Ans, Direction),
    search_answer(conds,     Ans, Conds),
    search_answer(condlist,  Ans, CondList),

    less_verbose(msg(['use: trying ', Name, ' (', Direction,
		') with conditions: ', Conds])),
    satisfy(Conds, CondList, ConjList),
    less_verbose(msg(['use: calling ', 
		trans(Name,Direction,Goal,CondList)])),
    trans(Name, Direction, Goal, CondList),

    %% Now write the explanatory msg.
    ((ConjList == [] ->
	MsgTail = [')'])
    ;
	(MsgTail = [', ', ConjList, ')'])
    ),
    msg(Msg).


%% discharge_all(+ConjectureList, +Tactic).
%%
%%   Applies Tactic to each of the conjectures in ConjectureList.
%%   It fails if any of the conjectures are not discharged.
%%   Therefore, Tactic should normally not involve user interaction,
%%   as the interactive proof steps will be lost if following
%%   conjectures are not discharged.

discharge_all([], _).

discharge_all([c(N)|CS], Tac) :-
	!,
	discharge_hyp(N, Tac),
	!,
	discharge_all(CS, Tac).


/*
 * print(Query).
 * print(Query, Parts).
 *
 * Searches for and prints all axioms/theorems that satisfy Query.
 */

print(Query,Parts) :-
    search_one_dirn(Query and print(Parts), _),
    fail.

print(_,_).


print(Query) :-
    search_one_dirn(Query and print, _),
    fail.

print(_).



%% fwd(+Query, ?Goal, ?ConjectureList).
%% fwd(+Query, ?Goal, ?ConjectureList).
%%
%%   This attempts to add Goal as a new hypothesis, using a rule 
%%   that satisfies Query and has Goal as its output.  The conditions
%%   and input of the rule are satisfied using existing hypotheses
%%   or theorems, or by adding conjectures if that is allowed by
%%   ConjectureList.  ConjectureList is instantiated to the list
%%   of conjectures added by this command.  Eg. [c(5),c(6)].

fwd(Q,G)   :- fwd(Q, G, []).

fwd(Query, GoalSpec, ConjList) :-
    srch_expand_term(GoalSpec, Goal),
    approx_copy(Goal, SearchGoal),
    ((var(Goal),(\+ frozen(Goal)) ->
	Msg = ['fwd(', Name ,', _' | MsgTail],
	OutputQuery = not(output = true))
    ;
	(Msg = ['fwd(', Name, ', ', Goal | MsgTail],
	 OutputQuery = (output = SearchGoal))
    ),
    check_or_writelist((nonvar(Query);nonvar(SearchGoal)),
	['fwd: query and/or goal arguments are too vague']),

    thawing_new_vars((
	search(Query and reln=(=>) and OutputQuery, Ans),
	search_answer(body,  Ans, Body),
	with_frozen_nfi_test(on,Body)
        )),
    search_answer(name,      Ans, Name),
    search_answer(direction, Ans, Direction),
    search_answer(conds,     Ans, Conds),
    search_answer(condlist,  Ans, CondList),
    search_answer(input,     Ans, Input),

    less_verbose(msg(['fwd: trying ', Name, ' (', Direction,
		') Input = ',Input,',  Conds = ', Conds])),
    ((ConjList == [] ->
	ConjList1 = [],   % this just prevents further searching as
	ConjList2 = [])   % soon as one condition cannot be satisfied.
    ;
	true
    ),	
    satisfy(Conds, CondList, ConjList1),
    fwd_satisfy_input(Input, Num, ConjList2, Safe),
    append(ConjList1, ConjList2, ConjList),

    less_verbose(msg(['fwd: calling ', 
		h_trans(Num,Name,Direction,Goal,CondList)])),

    %% compiled rules don't work with the 'rt' direction,
    %% so we generate a conjecture, then use the 't' direction.
    ((direction = rt, search_answer(kind, Ans, compiled_rule) ->
	conjecture(Goal, GNum),
	d_trans_true(GNum, Name, CondList))
    ;
	((Safe == yes ->
	    safe(h_trans(Num, Name, Direction, Goal, CondList)))
	;
	    h_trans(Num, Name, Direction, Goal, CondList)
	)
    ),

    %% Now write the explanatory msg.
    ((ConjList == [] ->
	MsgTail = [')'])
    ;
	(MsgTail = [', ', ConjList, ')'])
    ),
    msg(Msg).


%% fwd_satisfy_input(+Term, -HypNum, -ConjectureList, -Safe)
%%
%%  This is an auxillary procedure of fwd/3.  It ensures that the
%%  'input' of the rule to be used by fwd is available as a hypothesis
%%  and returns the number of that hypothesis in HypNum.
%%  Term is the required input.
%%  The Safe output is 'yes' if the fwd command should not delete
%%  HypNum, or 'no' if it can (because it has just been created
%%  by fwd_satisfy_input.

fwd_satisfy_input(true, true, [], no) :-
	!.

fwd_satisfy_input(Term, Num, [], yes) :-
	is_hyp(Num, Term),
	!.

fwd_satisfy_input(Term, Num, [], no) :-
	is_theorem2(Thm, Term),
	!,
	% do a take_theorem, returning Num.
	h_trans(true, Thm, rt, _, [], Num).

fwd_satisfy_input(Term, Num, [c(Num)], no) :-
	conjecture(Term, Num),
	!.


%% approx_copy(+Term0, -Term)
%%
%%   This constructs a copy of Term0 and returns it in Term.
%%   The resulting Term has the same structure as Term0,
%%   but non-frozen variables and terms involving substitutions 
%%   are replaced by anonymous variables.  The rationale for
%%   this is that we want to copy Term0, but do not want to
%%   have to reproduce all the nfi constraints on the variables
%%   in Term0.

approx_copy(In, Out) :-
	( %% we put the functor case first because it pushes
	  %% substitutions inside the functor.
	  (functor(In, Func, Arity) ->
	    functor(Out, Func, Arity),
	    approx_copy_args(In, Out, Arity))

	  %% this case must come next, so that all
	  %% substitutions are turned into anonymous variables.
	; (substitute(In, [_|_], Term) ->
	    Out = _)

	; (quantify(In, Q, _, _) ->
	    quantify(In, Q, x, Body),
	    approx_copy(Body, BodyOut),
	    quantify(Out, Q, x, BodyOut))

	; (frozen(In) ->
	    Out = In)
	;
	    true   % Out unconstrained.
	).


%% This copies the arguments of In.
%%
approx_copy_args(In, Out, N) :-
	(
	  (N > 0
	  -> arg(N, In, Arg),
	     arg(N, Out, ArgOut),
	     approx_copy(Arg, ArgOut),
	     N1 is N - 1,
	     approx_copy_args(In, Out, N1)
	  )
	;
	  true
	).
