%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS rewrite.ql,v 1.7 1993/12/10 02:19:04 marku Exp


?- op(810,xfy,and).


%% rewrite(+RuleList).
%%
%%  This version of rewrite is slow, because it tries each
%%  one of the rules in RuleList without checking if it matches
%%  the focus first.  However, it is very flexible, because
%%  each element of RuleList can be an arbitrary query.
%%  For better efficiency, use the rewrite_group tactic below.

:- tactic(rewrite(RuleList),
	['rewrite(+RuleList) -- rewrite the focus using the',nl,
	 'rules in RuleList.  Each element of RuleList can be',nl,
	 'a rule name or a more complex query (see use/1)'],
	['Rewriting using rules: ', RuleList],
	[]
    ).


%% rewrite_group(+RuleGroups).
%%
%%  This tactic is similar to, but more efficient than, the rewrite tactic.
%%  However, it requires you to set up an appropriate rule group beforehand
%%  (using the rule_group command).  Can take one group name or a list
%%  thereof.

:- tactic(rewrite_group(RuleGroups),
	['rewrite_group(+RuleGroups) -- rewrites the focus using',nl,
	 'the rules in the predefined rule groups RuleGroups.',nl,
	 'It also splits conjunctions, uses hypotheses and',nl,
	 'quits when the focus matches the goal'],
	['Rewriting using rule groups: ', RuleGroups],
	[]
    ).



%% rewrite_subterm(+RuleGroups).
%%
%%  This tactic will search for subterms at which to apply rules.
%%  However, it requires you to set up appropriate rule groups beforehand
%%  (using the rule_group command).

:- tactic(rewrite_subterm(RuleGroups),
	['rewrite_subterm(+RuleGroups) -- rewrites subterms of the focus',nl,
	 'using the rules in the predefined rule groups RuleGroups.',nl,
	 'RuleGroups can be a list of rule group names or a single name'],
	['Rewriting subterms using rule groups: ', RuleGroups],
	[]
    ).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

rewrite(Rules) :-
	isList(Rules),
	one_or_more_times(rewrite1(Rules)).

rewrite1(Rules) :-
	elem(R, Rules),
	use(R).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% rewrite_group(+Groups)

%% quit if focus==goal.
rewrite_group(_) :-
	is_focus(F),
	is_goal(G),
	F == G,
	complete_retry_delays,
	!.

%% split conjuncts.
rewrite_group(Groups) :-
	is_focus(_ and _),
	is_goal(true),
	!,
	((focus([1], empty, rewrite_group(Groups)) ->
	    Next = ok)
        ; 
	    true
	),
	((focus([2], empty, rewrite_group(Groups)) ->
	    Next = ok)
        ; 
	    true
	),
	nonvar(Next),
	((is_focus(true and _) ->
	    trans(r:trueand))
	;(is_focus(_ and true) ->
	    trans(r:andtrue))
	;
	    true
	).


%% quit after using a hypothesis to prove the focus.
rewrite_group(_) :-
	is_goal(true),
	is_focus(F),
	is_hyp(Num,F),
	trans_true(h(Num)),
	!.

%% try rules in the given RuleGroups.
rewrite_group(Groups) :-
	is_focus(F),
        elem(Group, Groups),
	get_up(rule_group, [Group, F, Conds, Name, Dir]),
	((Conds == true ->
	    CondList = [])
	;
	    satisfy(Conds, CondList, [])
	),
	trans(Name, Dir, _, CondList),
	msg([Name,' (',Dir,') on focus: ',F]),
	!,
	once((
	        rewrite_group(Groups)
	    ;
	        true
	    )).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% rewrite_subterm(+RuleGroups)
%%

rewrite_subterm(Groups) :-
        elem(Group, Groups),
        get_up(rule_group, [Group, F, Conds, Name, Dir]),
        is_focus_subterm(F, Path),
        focus(Path,
	     (((Conds == true ->
		   CondList = [])
	      ;
	         satisfy(Conds, CondList, [])
	      ),
	      trans(Name, Dir, _, CondList)
	     )),
	msg([Name,' (',Dir,') on focus: ',F]),
        !,
        once((
                rewrite_subterm(Groups)
             ;
                true
            )).


one_or_more_times(Code) :-
	call(Code),
	!,
	zero_or_more_times(Code).


zero_or_more_times(Code) :-
	call(Code),
	!,
	zero_or_more_times(Code).

zero_or_more_times(_).
	
