%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS search.ql,v 1.15 1993/12/10 02:19:07 marku Exp


/*
 * A tactic for searching for rules that satisfy a query.
 * ======================================================
 *
 * This module defines the following predicates:
 *
 *	search(Query, Answer)
 *	search_one_dirn(Query, Answer)
 *	search_answer(Selector, Answer, Part)
 *
 * plus a number of internal predicates that should not
 * be used outside this file.  All their names start with srch_.
 *
 * In addition, it defines the following predicates that are
 * not directly search-related, so may be useful in other modules.
 *
 *	occurs_in(SubTerm, Term)
 *
 * KNOWN BUGS: conditions like name = h Num may not work if they
 *             are nested inside 'or' or 'not'.
 */


/*
 * search(+Query, -Answer)
 * =======================
 *
 * This predicate searches the database for rules (ie. axioms,
 * definitions and theorems) that satisfy the given query. 
 *
 * Query is a boolean expression constructed using 'and,' 'or' and 'not'
 * (plus bracketing if necessary) from the following conditions:
 *
 *	Condition ::=
 *		  theory    =               Atom
 *		| name      =               [TheoryList :] Term
 *		| name      contains        Atom
 *		| direction =               (f|r|t|rt)
 *		| condlist  =               ListofRuleorHypNames
 *		| Part      (=|*=|contains) TermSpec
 *		| reln      =               Atom
 *		| print
 *		| print(Part)
 *		| print(List_of_Parts)
 *		| confirm
 *
 *	Part ::= 
 *		  stmt		(the entire statement of the rule)
 *		| input 	(one side of the conclusion)
 *		| output	(other side of the conclusion)
 *		| cond		(any condition of the rule)
 *		| conds		(the conjunction of all conditions)
 *
 *	TermSpec ::= focus | goal | h Patt_or_Num | PrologTerm
 *
 * Answer is a structure that contains lots of information
 * about a rule from the database that satisfies the Query.
 * This information should be accessed via the following
 * search_* access predicates.
 *
 * NB. The theory tests and some other conditions are executed more
 *     efficiently if they appear at the outermost level of conjunctions.
 */

?- op(700,xfx,(contains)).
?- op(700,xfx,(*=)).

?- op(800,fy,(not)).
?- op(810,xfy,(and)).
?- op(820,xfy,(or)).
?- op(830,xfy,(=>)).
?- op(840,xfy,(<=>)).



/* Predicates for accessing search Answers
 * =======================================
 */ 
search_answer(theory,	 ans(T,_,_,_,_,_,_,_,_,_,_,_,_,_,_,_), V) :- !,V=T.
search_answer(name,      ans(T,N,_,_,_,_,_,_,_,_,_,_,_,_,_,_), Name) :-
    !, ( (T==hyps -> Name = N) ; (Name = (T:N)) ).
search_answer(name2,	 ans(_,N,_,_,_,_,_,_,_,_,_,_,_,_,_,_), V) :- !,V=N.
search_answer(hypnum,	 ans(_,_,N,_,_,_,_,_,_,_,_,_,_,_,_,_), V) :- !,V=N.
search_answer(stmt,	 ans(_,_,_,S,_,_,_,_,_,_,_,_,_,_,_,_), V) :- !,V=S.
search_answer(body,	 ans(_,_,_,_,B,_,_,_,_,_,_,_,_,_,_,_), V) :- !,V=B.
search_answer(direction, ans(_,_,_,_,_,D,_,_,_,_,_,_,_,_,_,_), V) :- !,V=D.
search_answer(conds,	 ans(_,_,_,_,_,_,H,_,_,_,_,_,_,_,_,_), V) :- !,V=H.
search_answer(input,	 ans(_,_,_,_,_,_,_,I,_,_,_,_,_,_,_,_), V) :- !,V=I.
search_answer(reln,	 ans(_,_,_,_,_,_,_,_,R,_,_,_,_,_,_,_), V) :- !,V=R.
search_answer(output,	 ans(_,_,_,_,_,_,_,_,_,O,_,_,_,_,_,_), V) :- !,V=O.
search_answer(kind,      ans(_,_,_,_,_,_,_,_,_,_,_,K,_,_,_,_), V) :- !,V=K.
search_answer(unified,   ans(_,_,_,_,_,_,_,_,_,_,_,_,U,_,_,_), V) :- !,V=U.
search_answer(condlist,  ans(_,_,_,_,_,_,_,_,_,_,_,_,_,A,_,_), V) :- !,V=A.
search_answer(frozen,	 ans(_,_,_,_,_,_,_,_,_,_,_,_,_,_,_,F), V) :-
    nonvar(F),
    !,
    V = F.
search_answer(frozen, ans(hyps,_,_,F,_,_,_,_,_,_,_,_,_,_,_,F), V) :-
    !,
    V = F.
search_answer(frozen,	 ans(T,N,_,_,_,_,_,_,_,_,_,K,_,_,_,F), F) :-
    !,
    ((K == compiled_rule ->
	get_compiled_rule(T, N, _, F))
    ;
	find_rule_at2(N,F,_,T)
    ),
    freeze(F).   % freeze the new copy of the statement.

search_answer(T, _, _) :-
    !,
    write('Error: bad first argument to search_answer: '),
    writeln(T),
    fail.
    


search(Query, Answer) :-
    srch_1(Query, Answer, FCode, Concl),
    srch_2(Answer, FCode, Concl).


%% search_one_dirn(+Query, -Answer)
%%
%%   Same as search/2, except that it returns only the first
%%   direction of each rule that satisfies Query.

search_one_dirn(Query, Answer) :-
    srch_1(Query, Answer, FCode, Concl),
    once(srch_2(Answer, FCode, Concl)).



%% srch_1 -- first part of the search: finding a rule
%%
srch_1(Query, Answer, FCode, Concl) :-
    Answer = ans(Thy,Name,HypNum,Stmt,Body,
		 Dirn,Conds,Input,Reln,Output,
		 UOutput,Kind,Unified,AntList,_,_),
    srch_compile(Query, nec, true, Answer,
	      TCode, SCode, FCode, opt(TList,_)),
    !,
    current_dir(CD),
    (
	(var(TList)
        ->
	TCode2 = (((Thy=hyps);ancestor_of(CD,Thy)), once(TCode))
        )
    ;
	(TCode2 = (member(Thy,TList),once(TCode)))
    ),
    !,
    %% try each theory
    call(TCode2),
    ((Thy == hyps, in_proof ->
	    Name = h(_),
	    is_hyp(Name,HypNum,Stmt),
	    Body = true,
	    Kind = hypothesis)
    ;
	(member(Kind, [theorem, axiom, lemma]),
	 clause(([Kind,Thy] : [Name, Stmt]), Body))
    ;
	(Kind = compiled_rule,
	 get_compiled_rule(Thy, Name, _, Concl),
	 Body = true)
    ),
    once(call(SCode)).


%% srch_2 -- second part of the search: choosing a direction
%%
srch_2(Answer, FCode, Concl) :-
    Answer = ans(Thy,Name,_HypNum,Stmt,_Body,
		 Dirn,Conds,Input,Reln,Output,
		 _,Kind,Unified,AntList,_,_),

    %% try each direction
    choose_conditional(AntList, Stmt, Conds, Concl),
    choose_direction(Dirn, Concl, Input, Reln, Output),

    %% Check that Reln is a legal preorder.
    ((Reln = rev(R) ->
	preorder(R))
    ;
	preorder(Reln)
    ),

    %% execute the final query
    once(call(FCode)),

    %% evaluate compiled rules, if input/output/stmt has been instantiated.
    ((Kind = compiled_rule, nonvar(Unified) ->
	eval_compiled_rule(Thy, Name, _, Concl, Stmt, _))
    ;
	true
    ).
    


/*
 * srch_compile(Query, Nec, Truth, Answer,		(inputs)
 *		TheoryCode, StmtCode, FinalCode,	(output code)
 *		Optimise).				(other outputs)
 *
 *	Checks that Query is a legal search expression and, if so,
 *	compiles it into a series of three Prolog predicates
 *	(TheoryCode, StmtCode and FinalCode) which can be executed
 *	to determine whether or not the current instantiation of
 *	Answer satisfies Query.  These predicates rely on the 
 *	following components of Answer being instantiated at the 
 *	time that they are called:
 *
 *	    TheoryCode:	search_answer(theory, Answer).
 *
 *	    StmtCode:	the above, plus
 *			search_answer(nameX, Answer).
 *			search_answer(stmt, Answer).
 *
 *	    FinalCode:  all the above, plus
 *			search_answer(direction, Answer).
 *			search_answer(conds, Answer).
 *			search_answer(input, Answer).
 *			search_answer(reln, Answer).
 *			search_answer(output, Answer).
 *
 *	Note that the tests must be executed in the above order and
 *	must all be executed (since StmtCode may cache results that
 *	are needed by FinalCode, etc.).
 *
 *	The other arguments to srch_compile are:
 *
 *	    Nec    -	is the atom 'nec' if Query is an essential
 *			part of the total query, or 'maybe' otherwise
 *			(ie. if it is contained within a disjunction).
 *
 *	    Goal  -	The atom 'true' or the atom 'false'.
 *			If Nec == nec, then Goal determines whether
 *			it is necessary for Query to be true or false.
 *
 *	    Optimise -	Is a structure containing various optimisation 
 *			hints.  The variables within this structure may
 *			be instantiated by the compiler to indicate
 *			that various optimisations are possible.
 *			See srch_set_opt/3 for more details.
 */

/* This case must come first so that uninstantiated queries 
 * are instantiated correctly (as being always satisfied).
 */
srch_compile(Goal,_,Goal,_,Goal,Goal,Goal,_) :-
    !.

srch_compile(Q1 and Q2,Nec,Goal,Ans,TCode,SCode,FCode,Opt) :-
    !,
    (
	(Goal = true
	     ->
	 SubNec = Nec)
      ;
	(SubNec = maybe)
    ),
    srch_compile(Q1,SubNec,Goal,Ans,TCode1,SCode1,FCode1,Opt),
    !,
    srch_compile(Q2,SubNec,Goal,Ans,TCode2,SCode2,FCode2,Opt),
    srch_mk_and(TCode, TCode1, TCode2),
    srch_mk_and(SCode, SCode1, SCode2),
    srch_mk_and(FCode, FCode1, FCode2).
    %% TODO: optimise these, including simplifying FCode
    %%    to Goal if FCode == SCode and Nec == nec. etc.

srch_compile(Q1 or Q2,Nec,Goal,Ans,TCode,SCode,FCode,Opt) :-
    !,
    (
	(Goal = false
	     ->
	 SubNec = Nec)
      ;
	(SubNec = maybe)
    ),
    srch_compile(Q1,SubNec,Goal,Ans,TCode1,SCode1,FCode1,Opt),
    !,
    srch_compile(Q2,SubNec,Goal,Ans,TCode2,SCode2,FCode2,Opt),
    !,
    srch_mk_or(TCode, TCode1, TCode2),
    srch_mk_or(SCode, SCode1, SCode2),
    srch_mk_or(FCode, FCode1, FCode2).

srch_compile(not(Q1),Nec,Goal,Ans,TCode,SCode,FCode,Opt) :-
    !,
    srch_negate_goal(Goal,NotGoal),
    srch_compile(Q1,Nec,NotGoal,Ans,TCode1,SCode1,FCode1,Opt),
    srch_mk_not(TCode, TCode1),
    srch_mk_not(SCode, SCode1),
    srch_mk_not(FCode, FCode1).

/* two special forms for names with theories */
srch_compile(name = E, Nec, Goal, Ans, TCode, SCode, FCode, Opt) :-
    nonvar(E),
    E = T:N,
    !,
    srch_compile(theory = T and name = N, 
	Nec, Goal, Ans, TCode, SCode, FCode, Opt).

srch_compile(T:N, Nec, Goal, Ans, TCode, SCode, FCode, Opt) :-
    !,
    srch_compile(theory = T and name = N, 
	Nec, Goal, Ans, TCode, SCode, FCode, Opt).

srch_compile(Q, Nec, Goal, Ans, TCode, SCode, FCode, Opt) :-
    %% Q is a single condition.
    srch_compile_cond(Q, Nec, Goal, Ans, When, Code, Opt),
    srch_when(When, Code, Nec, Goal, TCode, SCode, FCode).



/*
 * Predicates used by srch_compile to construct and simplify code.
 *
 * srch_mk_and( AandB_code, A_code, B_code ).
 * srch_mk_or(  AorB_code,  A_code, B_code ).
 * srch_mk_not( NotA_code,  A_code ).
 */

srch_mk_and( B,     true,  B    ) :- !.
srch_mk_and( A,     A,     true ) :- !.
srch_mk_and( false, false, _    ) :- !.
srch_mk_and( false, _,     false) :- !.
srch_mk_and( (A,B), A,     B    ).

srch_mk_or( true,  true,  _     ) :- !.
srch_mk_or( true,  _,     true  ) :- !.
srch_mk_or( A,     A,     false ) :- !.
srch_mk_or( B,     false, B     ) :- !.
srch_mk_or( (A;B), A,     B     ).

srch_mk_not( false,  true  ) :- !.
srch_mk_not( true,   false ) :- !.
srch_mk_not( (\+ C), C     ).    



/*
 * srch_negate_goal(CurrentGoal, NegatedGoal).
 *
 *	Translates true to false and vice versa.
 */

srch_negate_goal(true, false).
srch_negate_goal(false, true).


/*
 * srch_when(When, Code, Nec, Goal, TCode, SCode, FCode)
 *
 *  This predicate determines when Code should be executed
 *  (after choosing the 'theory', the 'stmt', or the 'direction'),
 *  and whether or not that test needs to be repeated at later
 *  stages (eg. if it may interact with other tests).
 *
 *  NB. The current algorithm is unnecessarily conservative for
 *      queries like (Q1 or Q2) where Q1 and Q2 produce
 *      When=stmt or When=theory.  Q1 and Q2 will be repeated
 *      unnecessarily after the direction is chosen.  However,
 *      it is probably not worth the bother of fixing this.
 */

srch_when(direction, Code, _,     G, G,    G,    Code).
srch_when(stmt,      Code, nec,   G, G,    Code, G   ).
srch_when(stmt,      Code, maybe, G, G,    Code, Code).
srch_when(theory,    Code, nec,   G, Code, G,    G   ).
srch_when(theory,    Code, maybe, _, Code, Code, Code).

    

/* Theory conditions */
srch_compile_cond(theory = E,Nec,Goal,Ans,theory,Code,Opt) :-
    (var(E) ; atom(E)),
    !,
    search_answer(theory, Ans, T),
    Code = (T == E),
    srch_set_opt(theorylist, [E], Nec, Goal, Opt).

srch_compile_cond(theory = E, Nec,Goal,Ans,theory,Code,Opt) :-
    nonvar(E),
    E = [A|_],
    atom(A),
    !,
    search_answer(theory, Ans, T),
    Code = member(T,E),
    srch_set_opt(theorylist, E, Nec, Goal, Opt).


/* Name conditions */
srch_compile_cond(name = E, nec,true,Ans,stmt,Code,_) :-
    (var(E) ; functor(E,_,_)),
    !,
    search_answer(name2, Ans, E),
    Code = true.  % test is implemented by instantiating Ans.

srch_compile_cond(name = E, _,_,Ans,stmt,Code,_) :-
    (var(E) ; functor(E,_,_)),
    /* not( Nec==nec, Goal==true ) */
    /* not( E = (_ : _) */
    !,
    search_answer(name2, Ans, N),
    Code = (N = E).

srch_compile_cond(name contains E, _,_,Ans,stmt,Code,_) :-
    atom(E),
    !,
    search_answer(name2, Ans, N),
    search_answer(theory, Ans, T),
    atom_chars(E, EStr),
    Code = (\+ T=hyps, once(srch_name_contains(N,EStr))).


/* Direction conditions */
srch_compile_cond(direction = E, nec,true,Ans,direction,Code,_) :-
    nonvar(E),
    member(E, [f,r,t,rt]),
    !,
    search_answer(direction, Ans, E),
    Code = true.  % test is implemented by instantiating Ans.

srch_compile_cond(direction = E, _,_,Ans,direction,Code,_) :-
    /* not( Nec==nec, Goal==true ) */
    nonvar(E),
    member(E, [f,r,t,rt]),
    !,
    search_answer(direction, Ans, D),
    Code = (D == E).

srch_compile_cond(condlist = E, nec,true,Ans,direction,Code,_) :-
    isList(E),
    !,
    search_answer(condlist, Ans, E),
    Code = true.  % test is implemented by instantiating Ans.

srch_compile_cond(condlist = E, _,_,Ans,direction,Code,_) :-
    isList(E),
    !,
    search_answer(condlist, Ans, D),
    Code = (D = E).



/* Stmt (and subparts thereof) conditions */
srch_compile_cond(Part = E,  nec,true,Ans,direction,Code,_) :-
    member(Part, [input,output,stmt]),
    !,
    srch_expand_term(E, NewE),
    search_answer(Part, Ans, NewE),
    search_answer(unified, Ans, true),
    Code = true.  % test is implemented by instantiating Ans.

srch_compile_cond(cond = E, _,_,Ans,direction,Code,_) :-
    !,
    srch_expand_term(E, NewE),
    search_answer(conds, Ans, Conds),
    Code = (srch_choose_cond(Conds,H), H = NewE).

%% This one is not documented yet.
srch_compile_cond(conds = E, _,_,Ans,direction,Code,_) :-
    !,
    srch_expand_term(E, NewE),
    search_answer(conds, Ans, Conds),
    Code = (Conds = NewE).

srch_compile_cond(stmt = E,  _,_,Ans,stmt,Code,_) :-
    %% not(Nec=nec and Goal=true)
    !,
    srch_expand_term(E, NewE),
    search_answer(stmt, Ans, PartVar),
    search_answer(unified, Ans, Unified),
    Code = (PartVar = NewE, Unified = true).

srch_compile_cond(input = E,  _,_,Ans,direction,Code,_) :-
    %% not(Nec=nec and Goal=true)
    !,
    srch_expand_term(E, NewE),
    search_answer(input, Ans, PartVar),
    search_answer(unified, Ans, Unified),
    Code = (PartVar = NewE, Unified = true).

srch_compile_cond(output = E,  _,_,Ans,direction,Code,_) :-
    %% not(Nec=nec and Goal=true)
    !,
    srch_expand_term(E, NewE),
    search_answer(output, Ans, PartVar),
    search_answer(unified, Ans, Unified),
    Code = (PartVar = NewE, Unified = true).


srch_compile_cond(cond *= E, Nec,Goal,Ans,direction,Code,Opt) :-
    !,
    srch_expand_term(E, NewE),
    search_answer(condlist, Ans, AntList),
    Code = (search_answer(frozen, Ans, Stmt),
	    choose_conditional(AntList, Stmt, Conds, _),
	    srch_choose_cond(Conds,H),
	    H = NewE),
    %% The following two clauses attempt to optimise the search by
    %% limiting it to theories in which NewE is a legal term.
    once(term_theory(NewE, TList)),
    srch_set_opt(theorylist, TList, Nec, Goal, Opt).

srch_compile_cond(conds *= E, Nec,Goal,Ans,direction,Code,Opt) :-
    !,
    srch_expand_term(E, NewE),
    search_answer(condlist, Ans, AntList),
    Code = (search_answer(frozen, Ans, Stmt),
	    choose_conditional(AntList, Stmt, Conds, _),
	    Conds = NewE),
    %% The following two clauses attempt to optimise the search by
    %% limiting it to theories in which NewE is a legal term.
    once(term_theory(NewE, TList)),
    srch_set_opt(theorylist, TList, Nec, Goal, Opt).

srch_compile_cond(stmt *= E,  Nec,Goal,Ans,stmt,Code,Opt) :-
    !,
    srch_expand_term(E, NewE),
    Code = (search_answer(frozen, Ans, Stmt),
	     Stmt = NewE),
    (
	(Nec = nec,
	 Goal = true
	     ->
	 search_answer(stmt, Ans, NewE))  % speeds up the search.
      ;
	true
    ),
    %% The following two clauses attempt to optimise the search by
    %% limiting it to theories in which NewE is a legal term.
    once(term_theory(NewE, TList)),
    srch_set_opt(theorylist, TList, Nec, Goal, Opt).

srch_compile_cond(input *= E,  Nec,Goal,Ans,direction,Code,Opt) :-
    !,
    srch_expand_term(E, NewE),
    search_answer(condlist, Ans, AntList),
    search_answer(direction, Ans, Dirn),
    Code = (search_answer(frozen, Ans, Stmt),
	    choose_conditional(AntList, Stmt, _, Concl),
	    choose_direction(Dirn, Concl, Input, _, _),
	    Input = NewE),
    (
	(Nec = nec,
	 Goal = true
	     ->
	 search_answer(input, Ans, NewE))  % speeds up the search.
      ;
	true
    ),
    %% The following two clauses attempt to optimise the search by
    %% limiting it to theories in which NewE is a legal term.
    once(term_theory(NewE, TList)),
    srch_set_opt(theorylist, TList, Nec, Goal, Opt).

srch_compile_cond(output *= E,  Nec,Goal,Ans,direction,Code,Opt) :-
    !,
    srch_expand_term(E, NewE),
    search_answer(direction, Ans, Dirn),
    search_answer(condlist, Ans, AntList),
    Code = (search_answer(frozen, Ans, Stmt),
	    choose_conditional(AntList, Stmt, _, Concl),
	    choose_direction(Dirn, Concl, _, _, Output),
	    Output = NewE),
    (
	(Nec = nec,
	 Goal = true
	     ->
	 search_answer(output, Ans, NewE))  % speeds up the search.
      ;
	true
    ),
    %% The following two clauses attempt to optimise the search by
    %% limiting it to theories in which NewE is a legal term.
    once(term_theory(NewE, TList)),
    srch_set_opt(theorylist, TList, Nec, Goal, Opt).


srch_compile_cond(cond contains E, Nec,Goal,Ans,direction,Code,Opt) :-
    nonvar(E),
    !,
    srch_expand_term(E, NewE),
    search_answer(condlist, Ans, AntList),
    Code = once((
               search_answer(frozen, Ans, Stmt),
	       choose_conditional(AntList, Stmt, Conds, _),
	       occurs_in(NewE, Conds)
	       )),
    %% The following two clauses attempt to optimise the search by
    %% limiting it to theories in which NewE is a legal term.
    once(term_theory(NewE, TList)),
    srch_set_opt(theorylist, TList, Nec, Goal, Opt).

srch_compile_cond(stmt contains E,  Nec,Goal,Ans,stmt,Code,Opt) :-
    nonvar(E),
    !,
    srch_expand_term(E, NewE),
    Code = once((
               search_answer(frozen, Ans, Stmt),
	       occurs_in(NewE, Stmt)
	       )),
    %% The following two clauses attempt to optimise the search by
    %% limiting it to theories in which NewE is a legal term.
    once(term_theory(NewE, TList)),
    srch_set_opt(theorylist, TList, Nec, Goal, Opt).

srch_compile_cond(input contains E,  Nec,Goal,Ans,direction,Code,Opt) :-
    nonvar(E),
    !,
    srch_expand_term(E, NewE),
    search_answer(direction, Ans, Dirn),
    search_answer(condlist, Ans, AntList),
    Code = once((
               search_answer(frozen, Ans, Stmt),
	       choose_conditional(AntList, Stmt, _, Concl),
	       choose_direction(Dirn, Concl, Input, _, _),
	       occurs_in(NewE, Input)
	       )),
    %% The following two clauses attempt to optimise the search by
    %% limiting it to theories in which NewE is a legal term.
    once(term_theory(NewE, TList)),
    srch_set_opt(theorylist, TList, Nec, Goal, Opt).


srch_compile_cond(output contains E,  Nec,Goal,Ans,direction,Code,Opt) :-
    nonvar(E),
    !,
    srch_expand_term(E, NewE),
    search_answer(direction, Ans, Dirn),
    search_answer(condlist, Ans, AntList),
    Code = once((
               search_answer(frozen, Ans, Stmt),
	       choose_conditional(AntList, Stmt, _, Concl),
	       choose_direction(Dirn, Concl, _, _, Output),
	       occurs_in(NewE, Output)
	       )),
    %% The following two clauses attempt to optimise the search by
    %% limiting it to theories in which NewE is a legal term.
    once(term_theory(NewE, TList)),
    srch_set_opt(theorylist, TList, Nec, Goal, Opt).



/* Equivalence condition */
srch_compile_cond(reln = E,   _,_,Ans,direction,Code,_) :-
    !,
    search_answer(reln, Ans, Reln),
    Code = stronger(Reln,E).



srch_compile_cond(print,  _,_,Ans,direction,Code,_) :-
    !,
    srch_default_write_format(Fmt),
    Code = (incomplete_retry_delays,
	    srch_write_parts(Fmt,Ans)).

srch_compile_cond(print(Parts),  _,_,Ans,direction,Code,_) :-
    (atom(Parts) ; isList(Parts)),
    !,
    Code = (incomplete_retry_delays,
	    srch_write_parts(Parts,Ans)).


srch_compile_cond(confirm,  _,_,Ans,direction,Code,_) :-
    !,
    srch_default_write_format(Fmt),
    Code = (incomplete_retry_delays,
	    srch_write_parts(Fmt,Ans),
	    question('Try this rule?')).

srch_compile_cond(confirm(Parts),  _,_,Ans,direction,Code,_) :-
    (atom(Parts) ; isList(Parts)),
    !,
    Code = (incomplete_retry_delays,
	    srch_write_parts(Parts,Ans),
	    question('Try this rule?')).


srch_compile_cond(true,   _,_,_,direction,Code,_) :-
    !,
    Code = true.


/* some abbreviations of 'name = Name' */
srch_compile_cond(Name, Nec,Truth,Ans,stmt,Code,_) :-
    %% Name is not any of the above forms
    (atom(Name) ; (Name = def(_)) ; (Name = h(_))),
    !,
    srch_compile_cond(name = Name, Nec,Truth,Ans,stmt,Code,_).

srch_compile_cond(Query, _,_,_,_,_,_) :-
    !,
    write('Illegal search condition: '), 
    writeR(Query,_),
    nl,
    fail.


/*
 * srch_default_write_format(PrintList).
 *
 * Determines the default printing style for the 
 * print/0 and confirm/0 conditions.
 */

srch_default_write_format(
    [name, ' (',direction,')',
     if(conds==true,'',[tab,'Conds: ',conds]),
     if(body==true,
	if(kind==compiled_rule, [nl,tab,tab,'<<compiled rule>>'], []),
	[nl,tab,tab,'Side Conditions: ',body]),
     nl,tab,tab,input,tab(3),reln,tab(3),output,
     nl,nl
    ]
    ).


/*
 * srch_write_parts(Part, Ans)
 *
 * Writes out one or more parts of the selected rule,
 * as determined by Part.  Part may be the name of a rule component
 * (eg. input, reln, conds) or any atom to be written literally,
 * or a list of the above possibilities.
 */

srch_write_parts([], _) :-
    !.

srch_write_parts([P|Ps], Ans) :-
    !,
    srch_write_parts(P, Ans),
    srch_write_parts(Ps, Ans).

srch_write_parts(if(Part==E,Then,Else), Ans) :-
    !,
    (
	(search_answer(Part,Ans,E2), E==E2 ->
	    srch_write_parts(Then,Ans))
    ;
	    srch_write_parts(Else,Ans)
    ).

% Names of the form T:N are treated specially to avoid spaces around the ':'.
srch_write_parts(name, Ans) :-
    search_answer(name, Ans, T:N),
    !,
    writeR(T,_),
    writeR(':',_),
    writeR(N,_).

srch_write_parts(Part, Ans) :-
    member(Part, [theory,name,hypnum,stmt,body,
	          direction,conds,condlist,kind,
		  input,reln,output,frozen]),
    !,
    search_answer(Part, Ans, E),
    writeR(E,_).

srch_write_parts(Part, _) :-
    writelist([Part]).


/*
 * srch_set_opt(FlagName, Value, Nec, Goal, Opt)
 *
 *   If Nec=nec and Goal=true, this tries to instantiate the 
 *   FlagName component of Opt to Value (but succeeds anyway
 *   if it cannot.
 */

srch_set_opt(theorylist, T, nec, true, opt(T,_)) :- !.

srch_set_opt(_, _, _, _, _).



/*
 * srch_expand_term(E, NewE)
 *
 *	This expands the TermSpec abbreviations that can appear on the
 *	right hand side of conditions like: Part = E.
 */

srch_expand_term(E, E) :-
    var(E),
    !.

srch_expand_term(focus, E) :-
    !,
    is_focus(E).

srch_expand_term(goal, E) :-
    !,
    is_goal(E).

srch_expand_term(h(Patt_or_Num), E) :-
    !,
    is_hyp(h(Patt_or_Num), _, E),
    !.

srch_expand_term(E, E).
    %% E is none of the above abbreviations



/*
 * srch_choose_cond(Conds, H)
 *
 *	If Conds is a left-assoc conjunction, H will be instantiated 
 *	to each of the conjunctions within Conds in turn.  If Conds is 
 *	not a conjunction, H is unified with Conds.
 */

srch_choose_cond(H, H) :-
    var(H),
    !.

srch_choose_cond(true, _) :-
    !,
    fail.

srch_choose_cond(H1 and _, H1).

srch_choose_cond(_ and Conds, H) :-
    !,
    srch_choose_cond(Conds, H).
	
srch_choose_cond(H, H).
    %% H is not a conjunction, a variable, or true



/*
 * srch_name_contains(Name, Str)
 *
 *	Succeeds if Str is a substring of Name (or if they are equal).
 *
 *	Names of the form def(N) match Str iff Str is a substring
 *	of "def" or of N.
 *
 *	Name of the form Func(...) match Str iff Str is a substring
 *	of Func.
 */

srch_name_contains(def(Name), Str) :-
    atom_chars(Name, NameStr),
    append(Str, _, Suffix),
    append(_, Suffix, NameStr),
    !.

srch_name_contains(Name, Str) :-
    /* not (Name = def(_)) */
    functor(Name, FuncName, _),
    atom_chars(FuncName, FuncStr),
    append(Str, _, Suffix),
    append(_, Suffix, FuncStr).


/*
 * occurs_in(Subterm,Term)
 *
 *	Unifies Subterm with any subterm of Term that is not a variable.
 */

occurs_in(S,Term) :-
    substitute(Term, Subs, B),
    \+ ( Subs == [] ),
    !,
    (
	occurs_in(S,Subs)
      ;
	occurs_in(S,B)
    ).

occurs_in(_,X) :-
    var(X),
    !,
    fail.

occurs_in(_,X) :-
    is_object_var(X),
    !,
    fail.

occurs_in(S, S) :-
    /* not var(S) */
    !.

occurs_in(S, Term) :-
    functor(Term, _, N),
    !,
    occurs_in_args(N, S, Term).

occurs_in(S, Term) :-
    quantify(Term,_,_,B),
    !,
    occurs_in(S,B).


occurs_in_args(N, S,Term) :-
    \+ (N =:= 0),
    arg(N, Term, T),
    (
	occurs_in(S, T)
      ;
	(N1 is N - 1,
	 occurs_in_args(N1, S, Term)
	)
    ).
